//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	delib_io_tcp.cpp
//	project: delib
//
//
//  (c) DEDITEC GmbH, 2013
//  web: http://www.deditec.de/
//  mail: vertrieb@deditec.de
//
// 
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// DEBUG OUTPUT 
#define DEBUG_OUTPUT		1				// Ausgabe aktivieren (nur fr diese CPP-Datei)
// ------------------------------------

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	13.11.2013	Start DEDITEC-History



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// DapiOpenModuleIntern_ETH
// DapiCloseModuleIntern_ETH
// ------------------------------------
// tcp_tx_open
// tcp_tx_send_rec_data
// tcp_tx_close
// tcp_rx_close_session
// tcp_tx_send_rec_data_big
// ------------------------------------
// DapiTCPSendCommand
// DapiTCPSendCommandBig
// DapiTCPSendSpecialCommand
// DapiTCPSpecialCmd
// ------------------------------------
// DapiTCPWriteByte
// DapiTCPWriteWord
// DapiTCPWriteLong
// DapiTCPWriteLongLong
// ------------------------------------
// DapiTCPReadByte
// DapiTCPReadWord
// DapiTCPReadLong
// DapiTCPReadLongLong
// DapiTCPWriteMultipleBytes
// DapiTCPReadMultipleBytes
// ------------------------------------
// DapiTCPReadFifo

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include "proj_config.h"

#ifdef __linux__
	// LINUX ---------------
	// LINUX ---------------
	// LINUX ---------------
	
	#include "delib_linux_project_include-eth.h"
	
	#include "library/vc/deditec_tcp_io.h"
	#include "library/embedded/dt_embedded_parameter_linux.h"
	#include "library/delib/delib.h"
	#include "library/delib/delib_intern.h"
#else
	// WINDOWS ---------------
	// WINDOWS ---------------
	// WINDOWS ---------------
	#include "library/vc/comport.h"
	#include "library/ftdi/ftd2xx.h"
	#include "library/delib/delib.h"
	#include "library/delib/delib_intern.h"
	#include "library/delib/delib_extended_functions.h"
	#include "delib_set_get_error.h"
	#include "library/vc/deditec_buff_utils.h"
	#include "library/vc/deditec_string_utils.h"
	#include "library/vc/deditec_tcp_io.h"
	#include "library/embedded/dt_embedded_parameter.h"
	#include "delib_open_close.h"
#endif

// ------------------------------------
// ------------------------------------
// Defines
// ------------------------------------
// ------------------------------------

#define DapiTCPSendCommandRetryMax		5

#define BUFF_TCP_RECV_SIZE 				16384  	// better don't touch, used by read() and recv()
#define TX_BUFFER_CP_LENGTH 			1000
#define TCP_RO_BIG_RX_HEADER_LENGTH 	9		// id(2) + answer(1) + ok(1) + job_id(1) + length(4)
#define TCP_RO_1_RX_HEADER_LENGTH 		7		// id(2) + answer(1) + ok(1) + job_id(1) + length(2)
#define TCP_RO_BIG_TX_HEADER_LENGTH 	8		// id(2) + ro-cmd(1) + job_id(1) + length(4)

// ------------------------------------
// ------------------------------------
// DELIB-DEBUG
// ------------------------------------
// ------------------------------------
//
// Function															Windows (DEBUGVIEW)			Linux
// -------------------------										----------------------		-------------------------
// DapiSetError(debug_msg)											immer						immer
// debug_error(debug_msg)											immer						immer
// debug_info(debug_msg)											immer						nur bei DEBUG_OUTPUT
// debug_print(debug_msg)											nur bei DEBUG_OUTPUT		nur bei DEBUG_OUTPUT

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_tx_open(ULONG handle, unsigned char * host_name, unsigned long host_name_size, unsigned long port_no)
{
	char debug_msg[200];
	unsigned long ret;
	DAPI_HANDLE *DapiHandle;

	DapiHandle = (DAPI_HANDLE*) handle;
	
	TCP_IO * tcp_params; // creates global pointer, pointing to a structure of type TCP_IO (see deditec_tcp_io.h)
	tcp_params = (TCP_IO *) malloc(sizeof(TCP_IO)); // allocate ram space for struct TCP_IO beginning at pointer

	if (tcp_params == 0)
	{
		// DEBUG	
		debug_set_function_name("tcp_tx_open");
		sprintf(debug_msg,"return value is %d dec ", -5);
		debug_error(debug_msg);
		// -----

		return -5; // error allocating memory, failed
	}

	if(buffcpy_n(tcp_params->host_name, IP_ADDR_SERVER_SIZE, (unsigned char *) host_name, host_name_size) < 0)
	{
		return -6; // dest_buffer too small
	}
	
	tcp_params->host_name_size = host_name_size;
	tcp_params->port_no = port_no;

	// DEBUG
	debug_set_function_name("tcp_tx_open");
	sprintf(debug_msg,"tcp_params = 0x%lx", (unsigned long) tcp_params);//Sonst gehts unter Linux nicht ....
	debug_info(debug_msg);
	sprintf(debug_msg,"tcp_params = 0x%x", (unsigned int) tcp_params);
	debug_info(debug_msg);
	sprintf(debug_msg,"using IP: %s", tcp_params->host_name);
	debug_info(debug_msg);
	sprintf(debug_msg,"using port: %d", (unsigned int) tcp_params->port_no);
	debug_info(debug_msg);
	// -----

	ret = tcp_client_connect(tcp_params, host_name, host_name_size, tcp_params->port_no, 1); // returns pointer to tcp_params or neg. error code

	if (ret==-1 || ret==-2 || ret==-3 || ret==-4 || ret==-5|| ret==-6) // socket error
	{
		// DEBUG
		debug_set_function_name("tcp_tx_open");
		sprintf(debug_msg, "socket error - return is %d dec", (unsigned int) ret);
		debug_error(debug_msg);
		// -----

		return ret;
	}

	// DEBUG
	debug_set_function_name("tcp_tx_open");
	sprintf(debug_msg,"tcp_params->tcp_handle_tx = %d dez", (unsigned int) tcp_params->tcp_handle_tx);
	debug_info(debug_msg);
	sprintf(debug_msg,"function end - return value: 0x%x", (unsigned int) ret);
	debug_info(debug_msg);
	// -----

	return (unsigned long) tcp_params; // returns pointer to tcp_params
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_tx_send_rec_data( ULONG handle, unsigned long tcp_handle_tx, unsigned long job_id, unsigned long   tcp_timeout,
									unsigned long command_type,
									unsigned char * tx_buffer, unsigned long tx_buffer_length,
									unsigned char * rx_buffer, unsigned long rx_buffer_length )
{
	unsigned char tx_buffer_cp[TX_BUFFER_CP_LENGTH];
	unsigned char *recv_buffer;
	unsigned long recv_buffer_size;
	unsigned long recv_error;
	unsigned long recv_pos;
	unsigned long amount_bytes_sent;
	signed   long amount_bytes_received;
	unsigned long tx_buffer_cnt;
	unsigned long ret;
	struct timeval time_tcp_timeout;// = { 0, 100*1000 };
	unsigned long ret_select_bind;
	fd_set set;
	char debug_msg[200];
	DAPI_HANDLE *DapiHandle;
	
	DapiHandle = (DAPI_HANDLE*) handle;


	// DEBUG
	debug_collect_init();
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "tcp_tx_send_rec_data START");
	debug_collect(debug_msg);
	sprintf(debug_msg, "tcp_tx_send_rec_data(handle=0x%x, tx_buff_len=%ddez, rx_buff_len=%ddez)", (unsigned int) tcp_handle_tx, (unsigned int) tx_buffer_length, (unsigned int) rx_buffer_length);
	debug_collect(debug_msg);
	// -----
	
	recv_buffer_size = TCP_RO_1_RX_HEADER_LENGTH + rx_buffer_length;
	recv_buffer = (unsigned char *) malloc(recv_buffer_size);

	if (recv_buffer == NULL)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");
		sprintf(debug_msg, "malloc() recv_buffer failed");
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		return -2;
	}

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "%lu bytes memory allocated for recv_buffer", recv_buffer_size);
	debug_collect(debug_msg);
	sprintf(debug_msg, "setting tcp_timeout * tcp_timeout=%ld", tcp_timeout);
	debug_collect(debug_msg);
	// -----
	
	#ifndef __linux__
		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
	
		// timeout wird nur bei windows gesetzt
		time_tcp_timeout.tv_sec = (tcp_timeout / 1000);
		time_tcp_timeout.tv_usec = (tcp_timeout % 1000) * 1000;
		
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");
		sprintf(debug_msg, "tcp_timeout has been set to %ld sec and %ld usec", time_tcp_timeout.tv_sec, time_tcp_timeout.tv_usec);
		debug_collect(debug_msg);
		// -----
		
		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
	#endif
		
	if (tcp_handle_tx < 0) 
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");

		sprintf(debug_msg, "tcp_handle_tx < 0");
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		free(recv_buffer);
		return -1; // return socket error
	}
	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "tcp_handle_tx = 0x%x", (unsigned int) tcp_handle_tx);

	debug_collect(debug_msg);
	// -----
	
	
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// SENDING MESSAGE
	
	
	// ---- creating message header (3 bytes for ID's, 1 byte for job_id, 2 bytes for message length-> 6 bytes total)
	tx_buffer_cnt = 0;

	// now 3 bytes for DEDITEC ID and the command-id
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_0;
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_1;
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) command_type;//DEDITEC_TCP_PACKET_ID_2_CMD_RO_1;

	// now first byte for job_id
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) job_id;

	// now 2 bytes for length of receive buffer
	tx_buffer_cp[tx_buffer_cnt++] = ((unsigned char) ((tx_buffer_length+6) >> 8) & 255); // rotate by 8 bits and mask upper 8 bits
	tx_buffer_cp[tx_buffer_cnt++] =  (unsigned char) ((tx_buffer_length+6) & 255);       // mask lower 8 bits

	ret = buffcpy_n( tx_buffer_cp + 6,  TX_BUFFER_CP_LENGTH - 6,
					 tx_buffer,         tx_buffer_length        );
					 
	if (ret != 0)	// error
	{	
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");
		sprintf(debug_msg, "buffer too small return is %d dec", -2);
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----

		free(recv_buffer);
		return -2;
	}

	tx_buffer_cnt += tx_buffer_length;
	amount_bytes_sent = send(tcp_handle_tx, (const char *) tx_buffer_cp, tx_buffer_cnt, 0);
	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");
	sprintf(debug_msg, "SENDING %ld bytes:", amount_bytes_sent);
	debug_collect(debug_msg);
	debug_hexdump_buffer_to_string(tx_buffer_cp, amount_bytes_sent, debug_msg, sizeof(debug_msg));
	debug_collect(debug_msg);
	// -----
	
	
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// RECEIVING MESSAGE
	
	
	amount_bytes_received = 0;
	
	#ifdef __linux__
		amount_bytes_received = read(tcp_handle_tx, recv_buffer + amount_bytes_received, recv_buffer_size);
	#else
		FD_ZERO(&set);

		// Den Socket, der Verbindungen annimmt hinzufgen
		FD_SET(tcp_handle_tx, &set);
		
		ret_select_bind = select(tcp_handle_tx , &set, NULL, NULL, &time_tcp_timeout);
		amount_bytes_received=-1;

		if (FD_ISSET(tcp_handle_tx, &set)) 
		{
			amount_bytes_received = recv(tcp_handle_tx, (char *) recv_buffer , recv_buffer_size, 0);
		}
	#endif

	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");

	sprintf(debug_msg, "RECEIVING %ld bytes", amount_bytes_received);
	debug_collect(debug_msg);

	debug_hexdump_buffer_to_string(recv_buffer, amount_bytes_received, debug_msg, sizeof(debug_msg));
	debug_collect(debug_msg);
	// -----
	
	// antwort berprfen
	recv_error = 0;
	recv_pos = 0;

	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_0)					++recv_error;
	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_1)					++recv_error;
	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_2_ANSWER_RO_1)		++recv_error;
	if (recv_buffer[recv_pos++] != 0x0)										++recv_error;
	if (recv_buffer[recv_pos++] != (job_id & 0xff))							++recv_error;

	if (recv_error != 0)	// header not ok = error
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");

		sprintf(debug_msg, "error while receiving message: header of rx_msg not ok! - recv_error=%d", (unsigned int) recv_error);
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----

		free(recv_buffer);
		return -3; // failed to receive message from server
	}
	else	// header ok
	{
		if(buffcpy_n(rx_buffer, rx_buffer_length, recv_buffer+TCP_RO_1_RX_HEADER_LENGTH, amount_bytes_received-TCP_RO_1_RX_HEADER_LENGTH) == -1) // -1 if rx_buffer too small
		{		
			// DEBUG
			debug_set_function_name("tcp_tx_send_rec_data");

			sprintf(debug_msg, "rx_buffer too small return is %d dec", -5);
			debug_collect(debug_msg);

			debug_collect_printall();
			// -----

			free(recv_buffer);
			return -5;
		}
	}

	if (amount_bytes_received < 0) 
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data");

		sprintf(debug_msg, "recv return <0 (ret-value = %d) - function end - return is -3", (unsigned int) amount_bytes_received);
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----

		free(recv_buffer);
		return -3; // failed to receive message from server
	}

	if (amount_bytes_received > 0)  // seems possible to receive an empty tcp-packet
	{
		ret = buffcpy_n(rx_buffer, rx_buffer_length, recv_buffer+TCP_RO_1_RX_HEADER_LENGTH, amount_bytes_received-TCP_RO_1_RX_HEADER_LENGTH); // -1 if rx_buffer too small

		if (ret < 0)
		{		
			// DEBUG
			debug_set_function_name("tcp_tx_send_rec_data");

			sprintf(debug_msg, "rx_buffer too small return is %d dec", -5);
			debug_collect(debug_msg);

			debug_collect_printall();
			// -----
			
			free(recv_buffer);
			return -5;
		}
	}

	free(recv_buffer);

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data");

	sprintf(debug_msg, "tcp_tx_send_rec_data END * return=%ld", (amount_bytes_received - TCP_RO_1_RX_HEADER_LENGTH) );
	debug_collect(debug_msg);
	// -----
	
	return (amount_bytes_received - TCP_RO_1_RX_HEADER_LENGTH);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_tx_close(unsigned long handle)
{
//	char debug_msg[200];
	unsigned int ret;
	TCP_IO * tcp_params;
	DAPI_HANDLE *DapiHandle;

	DapiHandle = (DAPI_HANDLE*) handle;

	tcp_params = (TCP_IO *) handle; // creates pointer, pointing to a structure of type TCP_IO (see tcp_io.h)
	ret = tcp_disconnect(tcp_params);
	free(tcp_params);

	// DEBUG
//	debug_set_function_name("tcp_tx_close");
//	sprintf(debug_msg,"function end - return is %d dec", ret);
//	debug_info(debug_msg);
	// -----

	return ret;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

#ifdef __linux__
   unsigned long tcp_rx_close_session(unsigned long socket_file_descr) //closing server session
   {
      return close(socket_file_descr);
   }
#endif

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long tcp_tx_send_rec_data_big(	ULONG handle, unsigned long   tcp_handle_tx, unsigned long job_id, unsigned long   tcp_timeout,
										unsigned long command_type,
										unsigned char * tx_buffer, unsigned long tx_buffer_length,
										unsigned char * rx_buffer, unsigned long rx_buffer_length )
{
	unsigned char *tx_buffer_cp;
	unsigned long tx_buffer_cp_size;
	unsigned char *recv_buffer;
	unsigned long recv_buffer_size;
	unsigned long recv_error;
	unsigned long recv_pos;
	unsigned long amount_bytes_sent;
	signed   long amount_bytes_received;
	unsigned long tx_buffer_cnt;
	unsigned long ret;
	struct timeval time_tcp_timeout;
	unsigned long ret_select_bind;
	fd_set set;
	char debug_msg[200];
	DAPI_HANDLE *DapiHandle;
	
	DapiHandle = (DAPI_HANDLE*) handle;

	
	// DEBUG
	debug_collect_init();
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "tcp_tx_send_rec_data_big START");
	debug_collect(debug_msg);
	sprintf(debug_msg, "tcp_tx_send_rec_data_big(handle=0x%x, tx_buff_len=%ddez, rx_buff_len=%ddez)", (unsigned int) tcp_handle_tx, (unsigned int) tx_buffer_length, (unsigned int) rx_buffer_length);
	debug_collect(debug_msg);
	// -----
	
	tx_buffer_cp_size = TCP_RO_BIG_TX_HEADER_LENGTH + tx_buffer_length;
	tx_buffer_cp = (unsigned char *) malloc(tx_buffer_cp_size);

	if (tx_buffer_cp == NULL)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "malloc() tx_buffer_cp failed");
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		return -2;
	}

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "%lu bytes memory allocated for tx_buffer_cp", tx_buffer_cp_size);
	debug_collect(debug_msg);
	// -----
	
	recv_buffer_size = TCP_RO_BIG_RX_HEADER_LENGTH + rx_buffer_length;
	recv_buffer = (unsigned char *) malloc(recv_buffer_size);

	if (recv_buffer == NULL)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "malloc() recv_buffer failed");
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		free(tx_buffer_cp);	
		return -2;
	}

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "%lu bytes memory allocated for recv_buffer", recv_buffer_size);
	debug_collect(debug_msg);
	// -----
	
	#ifndef __linux__
		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
	
		// timeout wird nur bei windows gesetzt
		time_tcp_timeout.tv_sec = (tcp_timeout / 1000);
		time_tcp_timeout.tv_usec = (tcp_timeout % 1000) * 1000;
		
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "tcp_timeout has been set to %ld sec and %ld usec", time_tcp_timeout.tv_sec, time_tcp_timeout.tv_usec);
		debug_collect(debug_msg);
		// -----
		
		// WINDOWS ---------------
		// WINDOWS ---------------
		// WINDOWS ---------------
	#endif
	
	if (tcp_handle_tx < 0)
	{	
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "tcp_handle_tx < 0");
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----
		
		free(tx_buffer_cp);
		free(recv_buffer);
		return -1; // return socket error
	}
	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "tcp_handle_tx = 0x%x", (unsigned int) tcp_handle_tx);
	debug_collect(debug_msg);
	// -----
	
	
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// SENDING MESSAGE
	
	
	// ---- creating message header (3 bytes for ID's, 1 byte for job_id, 4 bytes for message length-> 8 bytes total)
	tx_buffer_cnt = 0;

	// now 3 bytes for DEDITEC ID and the command-id
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_0;
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_1;
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) DEDITEC_TCP_PACKET_ID_2_CMD_RO_BIG;

	// now first byte for job_id
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) (job_id & 0xff);

	// now 4 bytes for length of receive buffer
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) ((recv_buffer_size >> 24) & 255);
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) ((recv_buffer_size >> 16) & 255);
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) ((recv_buffer_size >> 8 ) & 255);
	tx_buffer_cp[tx_buffer_cnt++] = (unsigned char) ((recv_buffer_size      ) & 255);

	ret = buffcpy_n( tx_buffer_cp + tx_buffer_cnt,  tx_buffer_cp_size - tx_buffer_cnt, 
					 tx_buffer, 					tx_buffer_length);

	if (ret != 0)
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");
		sprintf(debug_msg, "buffer too small return is %d dec", -2);
		debug_collect(debug_msg);
		debug_collect_printall();
		// -----

		free(tx_buffer_cp);
		free(recv_buffer);		
		return -2;
	}

	tx_buffer_cnt += tx_buffer_length;
	amount_bytes_sent = send(tcp_handle_tx, (const char *) tx_buffer_cp, tx_buffer_cnt, 0);

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");
	sprintf(debug_msg, "SENDING %ld bytes:", amount_bytes_sent);
	debug_collect(debug_msg);
	debug_hexdump_buffer_to_string(tx_buffer_cp, amount_bytes_sent, debug_msg, sizeof(debug_msg));
	debug_collect(debug_msg);
	// -----
	
	
	// ----------------------------------------------------------------------------
	// ----------------------------------------------------------------------------
	// RECEIVING MESSAGE
	
	
	amount_bytes_received = 0;

	#ifdef __linux__
		amount_bytes_received = read(tcp_handle_tx, recv_buffer, recv_buffer_size);
	#else
		FD_ZERO(&set);

		// Den Socket, der Verbindungen annimmt hinzufgen
		FD_SET(tcp_handle_tx, &set);
		
		ret_select_bind = select(tcp_handle_tx , &set, NULL, NULL, &time_tcp_timeout);
		amount_bytes_received=-1;

		if (FD_ISSET(tcp_handle_tx, &set)) 
		{
			amount_bytes_received = recv(tcp_handle_tx, (char *) recv_buffer , recv_buffer_size, 0);
		}
	#endif

	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");

	sprintf(debug_msg, "RECEIVING %ld bytes", amount_bytes_received);
	debug_collect(debug_msg);

	debug_hexdump_buffer_to_string(recv_buffer, amount_bytes_received, debug_msg, sizeof(debug_msg));
	debug_collect(debug_msg);
	// -----
	
	// antwort berprfen
	recv_error = 0;
	recv_pos = 0;

	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_0)					++recv_error;
	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_1)					++recv_error;
	if (recv_buffer[recv_pos++] != DEDITEC_TCP_PACKET_ID_2_ANSWER_RO_BIG)	++recv_error;
	if (recv_buffer[recv_pos++] != 0x0)										++recv_error;
	if (recv_buffer[recv_pos++] != (job_id & 0xff))							++recv_error;

	if (recv_error != 0)	// error
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");

		sprintf(debug_msg, "error while receiving message: header of rx_msg not ok! - recv_error=%d", (unsigned int) recv_error);
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----
		
		free(tx_buffer_cp);
		free(recv_buffer);
		return -3; // failed to receive message from server
	}
	else	// header ok
	{
		if(buffcpy_n(rx_buffer, rx_buffer_length, recv_buffer+TCP_RO_BIG_RX_HEADER_LENGTH, amount_bytes_received-TCP_RO_BIG_RX_HEADER_LENGTH) == -1) // -1 if rx_buffer too small
		{
			// DEBUG
			debug_set_function_name("tcp_tx_send_rec_data_big");

			sprintf(debug_msg, "rx_buffer too small return is %d dec", -5);
			debug_collect(debug_msg);

			debug_collect_printall();
			// -----
			
			free(recv_buffer);
			free(tx_buffer_cp);
			return -5;
		}
	}

	if (amount_bytes_received < 0) 
	{
		// DEBUG
		debug_set_function_name("tcp_tx_send_rec_data_big");

		sprintf(debug_msg, "recv return <0 (ret-value = %d) - function end - return is -3", (unsigned int) amount_bytes_received);
		debug_collect(debug_msg);

		debug_collect_printall();
		// -----

		free(tx_buffer_cp);
		free(recv_buffer);
		
		return -3; // failed to receive message from server
	}

	if (amount_bytes_received > 0)  // seems possible to receive an empty tcp-packet
	{		
		ret = buffcpy_n(rx_buffer, rx_buffer_length, recv_buffer+TCP_RO_BIG_RX_HEADER_LENGTH, amount_bytes_received-TCP_RO_BIG_RX_HEADER_LENGTH); // -1 if rx_buffer too small

		if (ret < 0)
		{
			// DEBUG
			debug_set_function_name("tcp_tx_send_rec_data_big");

			sprintf(debug_msg, "rx_buffer too small return is %d dec", -5);
			debug_collect(debug_msg);

			debug_collect_printall();
			// -----

			free(tx_buffer_cp);
			free(recv_buffer);
			return -5;
		}
	}

	free(tx_buffer_cp);
	free(recv_buffer);
	
	// DEBUG
	debug_set_function_name("tcp_tx_send_rec_data_big");

	sprintf(debug_msg, "tcp_tx_send_rec_data_big END * return=%ld", (amount_bytes_received - TCP_RO_1_RX_HEADER_LENGTH) );
	debug_collect(debug_msg);
	// -----
	
	return (amount_bytes_received - TCP_RO_BIG_RX_HEADER_LENGTH);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiOpenModuleIntern_ETH(ULONG handle, ULONG moduleID, ULONG moduleNr, unsigned char* exbuffer)
{
	char debug_msg[200];
	unsigned char ip_addr[IP_ADDR_SERVER_SIZE];
	ULONG ret;
	char reg_key_name[60];
	char reg_key_param[70];
	ULONG reg_key_value_port;
	ULONG reg_key_value_enabled;
	ULONG reg_key_value_timeout;
	ULONG err;

	DAPI_OPENMODULEEX_STRUCT* exbuffer_int;
	TCP_IO * tcp_params;
	DAPI_HANDLE *DapiHandle;

	DapiHandle = (DAPI_HANDLE*) handle;

	if (handle ==0)
	{
		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"Invalid Handle (handle=0)");
		debug_error(debug_msg);
		// -----

		return FALSE;
	}

	if (DapiHandle->id != delib_id)
	{
		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"Invalid Handle-ID");
		debug_error(debug_msg);
		// -----

		return FALSE;
	}

	if(DapiCheckHandle(handle) != 0) 
	{	
		return -1;
	}

	if(exbuffer == 0)
	{
		#ifndef __linux__
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// --------------------------------------------------------------------
			DapiInternGetRegistryKeyName(moduleID, (char*) reg_key_name, sizeof(reg_key_name));
			// --------------------------------------------------------------------
			//
			// ------------------------------------------------
			// Registry read Server enabled
			sprintf(reg_key_param, "Enabled%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_DWORD, reg_key_name, reg_key_param, 0, 0, &reg_key_value_enabled);
			if (err == RETURN_OK)
			{
				sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=%d\n", reg_key_name, reg_key_param, reg_key_value_enabled);
				debug_info(debug_msg);
			}
			else
			{
				sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
				debug_info(debug_msg);
			}
			// ------------------------------------------------
			//
			//
			//
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// ------------------------------------------------
			// Registry read Server Port-No
			sprintf(reg_key_param, "Port%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_DWORD, reg_key_name, reg_key_param, 0, 0, &reg_key_value_port);
			if (err == RETURN_OK)
			{
				sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=%d\n", reg_key_name, reg_key_param, reg_key_value_port);
				debug_info(debug_msg);
			}
			else
			{
				sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
				debug_info(debug_msg);
			}
			// ------------------------------------------------
			//
			//
			//
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// ------------------------------------------------
			// Registry read Server Timeout
			sprintf(reg_key_param, "Timeout%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_DWORD, reg_key_name, reg_key_param, 0, 0, &reg_key_value_timeout);
			if (err == RETURN_OK)
			{
				sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=%d\n", reg_key_name, reg_key_param, reg_key_value_timeout);
				debug_info(debug_msg);
			}
			else
			{
				sprintf(debug_msg,"GET_REGISTRY_DWORD: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
				debug_info(debug_msg);

				// default value
				reg_key_value_timeout = 5000;

				debug_error("Parameter not found in Registry: Using default Value: timeout=5000");
			}
			// ------------------------------------------------
			//
			//
			//
			debug_set_function_name("DapiOpenModuleIntern_ETH");
			// ------------------------------------------------
			// Registry read Server Hostname
			sprintf(reg_key_param, "Hostname%d", moduleNr);
			// ------------------------------------------------
			err = DapiInternCommand(DAPI_INTERN_CMD_GET_REGISTRY_STRING, reg_key_name, reg_key_param, (char*) ip_addr, IP_ADDR_SERVER_SIZE, (ULONG*) 0);
			if (err == RETURN_OK)
			{
				sprintf(debug_msg,"GET_REGISTRY_STRING: Key=\"%s\" * Parameter=\"%s\" - Registry-Value=\"%s\"\n", reg_key_name, reg_key_param, ip_addr);
				debug_info(debug_msg);
			}
			else
			{
				sprintf(debug_msg,"GET_REGISTRY_STRING: Key=\"%s\" * Parameter=\"%s\" - NOT FOUND IN REGISTRY\n", reg_key_name, reg_key_param);
				debug_info(debug_msg);
			}
			// ------------------------------------------------
		#endif
	}
	else
	{
		// exbuffer wird benutzt
		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"exbuffer read");
		debug_info(debug_msg);
		// -----

		exbuffer_int = (DAPI_OPENMODULEEX_STRUCT*) exbuffer;

		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"ip=%s", exbuffer_int->address);						debug_info(debug_msg);
		sprintf(debug_msg,"port=%d", (unsigned int) exbuffer_int->portno);		debug_info(debug_msg);
		sprintf(debug_msg,"timeout=%ld", exbuffer_int->timeout);				debug_info(debug_msg);
		// -----

		reg_key_value_enabled 	= 1;
		reg_key_value_port 		= exbuffer_int->portno;
		reg_key_value_timeout 	= exbuffer_int->timeout;
		strcpy_n((char*) ip_addr, sizeof(ip_addr), (char*) exbuffer_int->address, sizeof(exbuffer_int->address));
	}

	if(reg_key_value_enabled != 1)
	{
		// Enabled in registry is not =1
		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"enabled !=1 -> return 0");;
		debug_info(debug_msg);
		// -----

		return -1;
	}
	else
	{
		if(reg_key_value_port==0)
		{
			// Default Port
			reg_key_value_port = 9912;//DEDITEC_DEFAULT_PORT_NO; //9912;//
		}
		
		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"trying to connect to server %s:%d", ip_addr, (unsigned int) reg_key_value_port);
		debug_info(debug_msg);
		// -----

		ret = tcp_tx_open(handle, ip_addr, sizeof(ip_addr), reg_key_value_port); //handle is false, it contains adr. of tcp_params

		// DEBUG
		debug_set_function_name("DapiOpenModuleIntern_ETH");
		sprintf(debug_msg,"tcp_tx_open returns %d dez", (unsigned int) ret);
		debug_info(debug_msg);
		// -----

		if (ret==-1 || ret==-2 || ret==-3 || ret==-4 || ret==-5|| ret==-6) // socket error
		{
			// Connect Error
			DapiHandle->TCPHandle = 0;	// 14.2.2012
			DapiHandle->moduleID = -1;
			DapiHandle->moduleNr = -1;
			return -1;
		}
		else
		{
			DapiHandle->TCPHandle = ret;
			tcp_params = (TCP_IO*) ret;
			DapiHandle->TCPHandle_tx = tcp_params->tcp_handle_tx;
			DapiHandle->moduleID = moduleID;
			DapiHandle->moduleNr = moduleNr;
			DapiHandle->SubDevice = 2;
			DapiHandle->timeout = reg_key_value_timeout;
		}
	}
	
	return 0;	// -> ok
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiCloseModuleIntern_ETH(ULONG handle)
{
//	char debug_msg[200];
	DAPI_HANDLE *DapiHandle;
	
   	DapiHandle = (DAPI_HANDLE*) handle;

/*	if (handle == 0)
	{
		// DEBUG
		debug_set_function_name("DapiCloseModuleIntern_ETH");
		sprintf(debug_msg, "Invalid Handle (handle=0)");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if (DapiHandle->id != delib_id)
	{
		// DEBUG
		debug_set_function_name("DapiCloseModuleIntern_ETH");
		sprintf(debug_msg, "Invalid Handle-ID");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if(DapiCheckHandle(handle) != 0) 
	{
		return -1;
	}
*/

	// "handle" ist ein gltiges Handle.....muss nicht mehr gecheckt werden
	tcp_tx_close(DapiHandle->TCPHandle);

	DapiHandle->TCPHandle = 0;
	DapiHandle->moduleID = -1;
	DapiHandle->moduleNr = -1;

	return TRUE;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

bool DapiTCPSendCommand(ULONG handle, UCHAR tx[], unsigned long tx_length, UCHAR rx[], unsigned long rx_anzahl)
{
	char debug_msg[200];
	unsigned long ret;
	unsigned long buff_cnt=0;
	unsigned char modul_nr = 0;
	
	DAPI_HANDLE *DapiHandle;
	
	DapiHandle = (DAPI_HANDLE*) handle;

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTCPSendCommand");
		sprintf(debug_msg,"tx_length=%d tx_buffer[0]=%c, tx_buffer[1]=%c rx_cnt=%d", (unsigned int) tx_length, tx[0],tx[1], (unsigned int) rx_anzahl);
		debug_print(debug_msg);
	#endif
	// -----
	

	if (handle == 0)
	{
		// DEBUG
		debug_set_function_name("DapiTCPSendCommand");
		sprintf(debug_msg, "Invalid Handle (handle=0)");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if (DapiHandle->id != delib_id)
	{
		// DEBUG
		debug_set_function_name("DapiTCPSendCommand");
		sprintf(debug_msg, "Invalid Handle-ID");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if(DapiCheckHandle(handle) != 0) 
	{
		return FALSE;
	}

	++DapiHandle->jobID;
	
	ret=tcp_tx_send_rec_data(handle, DapiHandle->TCPHandle_tx, DapiHandle->jobID, DapiHandle->timeout, DEDITEC_TCP_PACKET_ID_2_CMD_RO_1, tx, tx_length, rx, rx_anzahl);

	return TRUE;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

bool DapiTCPSendCommandBig(ULONG handle, UCHAR tx[], unsigned long tx_length, UCHAR rx[], unsigned long rx_anzahl, unsigned long *rx_data_length)
{
	char debug_msg[200];
	unsigned long buff_cnt = 0;
	unsigned char modul_nr = 0;
	DAPI_HANDLE *DapiHandle;
	
	DapiHandle = (DAPI_HANDLE*) handle;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTCPSendCommandBig");
		sprintf(debug_msg, "tx_length=%d tx_buffer[0]=%c, tx_buffer[1]=%c rx_cnt=%d", (unsigned int) tx_length, tx[0],tx[1], (unsigned int) rx_anzahl);
		debug_print(debug_msg);
	#endif
	// -----
		

	if (handle == 0)
	{
		// DEBUG
		debug_set_function_name("DapiTCPSendCommandBig");
		sprintf(debug_msg, "Invalid Handle (handle=0)");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if (DapiHandle->id != delib_id)
	{
		// DEBUG
		debug_set_function_name("DapiTCPSendCommandBig");
		sprintf(debug_msg, "Invalid Handle-ID");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if(DapiCheckHandle(handle) != 0) return FALSE;

	++DapiHandle->jobID;
		
	*rx_data_length = tcp_tx_send_rec_data_big(handle, DapiHandle->TCPHandle_tx, DapiHandle->jobID, DapiHandle->timeout, DEDITEC_TCP_PACKET_ID_2_CMD_RO_1, tx, tx_length, rx, rx_anzahl);

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTCPSendCommandBig");
		sprintf(debug_msg, "DapiTCPSendCommandBig ENDE");
		debug_print(debug_msg);
	#endif
	// -----

	return TRUE;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

bool DapiTCPSendSpecialCommand(ULONG handle, UCHAR tx[], unsigned long tx_length, UCHAR rx[], unsigned long rx_anzahl)
{
	char debug_msg[200];
	unsigned long ret;
	unsigned long retry = DapiTCPSendCommandRetryMax;	
	unsigned long buff_cnt = 0;
	unsigned char modul_nr = 0;
	DAPI_HANDLE *DapiHandle;	

	DapiHandle = (DAPI_HANDLE*) handle;
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTCPSendSpecialCommand");
		sprintf(debug_msg, "tx_length=%d tx_buffer[0]=%c, tx_buffer[1]=%c rx_cnt=%d", (unsigned int) tx_length, tx[0],tx[1], (unsigned int) rx_anzahl);
		debug_print(debug_msg);
	#endif
	// -----


	if (handle ==0)
	{
		// DEBUG
		debug_set_function_name("DapiTCPSendSpecialCommand");
		sprintf(debug_msg, "Invalid Handle (handle=0)");
		debug_error(debug_msg);
		// -----
	
		return FALSE;
	}

	if (DapiHandle->id != delib_id)
	{
		// DEBUG
		debug_set_function_name("DapiTCPSendSpecialCommand");
		sprintf(debug_msg, "Invalid Handle-ID");
		debug_error(debug_msg);
		// -----
		
		return FALSE;
	}

	if(DapiCheckHandle(handle) != 0) 
	{
		return FALSE;
	}

	++DapiHandle->jobID;

	ret=tcp_tx_send_rec_data(handle, DapiHandle->TCPHandle_tx, DapiHandle->jobID, DapiHandle->timeout, DEDITEC_TCP_PACKET_ID_2_CMD_SPECIAL, tx, tx_length, rx, rx_anzahl);

	return TRUE;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiTCPWriteByte(unsigned long handle, unsigned long adress, unsigned long value)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt=0;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='B';		// BYTE
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	tx_buffer[tx_cnt++]=(UCHAR) (value >> 0  ) & 0xff;

	if(!DapiTCPSendCommand(handle, tx_buffer, tx_cnt, 0, 0)) 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiTCPWriteWord (unsigned long handle,unsigned long adress,unsigned long value)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt=0;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='W';		// WORD
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;
	
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 8  ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 0  ) & 0xff;

	if(!DapiTCPSendCommand(handle, tx_buffer, tx_cnt, 0, 0)) 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiTCPWriteLong (unsigned long handle,unsigned long adress,unsigned long value)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt=0;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='L';		// Long
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	tx_buffer[tx_cnt++]=(UCHAR) (value >> 24 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 16 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 8  ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 0  ) & 0xff;

	if(!DapiTCPSendCommand(handle, tx_buffer, tx_cnt, 0, 0)) 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------


void DapiTCPWriteLongLong (unsigned long handle,unsigned long adress, ULONGLONG value)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt = 0;

	tx_buffer[tx_cnt++]='W';		// Write
	tx_buffer[tx_cnt++]='X';		// LongLong
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	tx_buffer[tx_cnt++]=(UCHAR) (value >> 56 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 48 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 40 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 32 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 24 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 16 ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 8  ) & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (value >> 0  ) & 0xff;

	if(!DapiTCPSendCommand(handle, tx_buffer, tx_cnt, 0, 0)) 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPReadByte (unsigned long handle,unsigned long adress)
{
	UCHAR tx_buffer[256];
	UCHAR rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='B';		// BYTE

	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	rx_cnt=1;		// 1 Byte soll empfangen werden

	if(DapiTCPSendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt)) 
	{
		data =((unsigned long) (rx[0]) << 0) & 0x00ff;
		return data;
	}
	else 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPReadWord (unsigned long handle,unsigned long adress)
{
	UCHAR tx_buffer[256];
	UCHAR rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;


	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='W';		// WORD
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	rx_cnt=2;		// 2 Bytes sollen empfangen werden

	if(DapiTCPSendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt)) 
	{
		data =((unsigned long) (rx[1]) << 8) & 0xff00;
		data|=((unsigned long) (rx[0]) << 0) & 0x00ff;
		return data;
	}
	else 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}
	
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

unsigned long DapiTCPReadLong(unsigned long handle, unsigned long adress)
{
	UCHAR tx_buffer[256];
	UCHAR rx[256];
	unsigned long data;
	unsigned long tx_cnt=0;
	unsigned long rx_cnt;


	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='L';		// LONG
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	rx_cnt=4;		// 4 Bytes sollen empfangen werden

	if(DapiTCPSendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt)) 
	{
		data =((unsigned long) (rx[3]) << 24) & 0xff000000;
		data|=((unsigned long) (rx[2]) << 16) & 0x00ff0000;
		data|=((unsigned long) (rx[1]) << 8 ) & 0x0000ff00;
		data|=((unsigned long) (rx[0]) << 0 ) & 0x000000ff;

		return data;
	}
	else 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONGLONG DapiTCPReadLongLong(unsigned long handle, unsigned long adress)
{
	UCHAR tx_buffer[256];
	UCHAR rx[256];
	ULONGLONG data;
	unsigned long tx_cnt = 0;
	unsigned long rx_cnt;

	tx_buffer[tx_cnt++]='R';		// Read
	tx_buffer[tx_cnt++]='X';		// LONGLONG
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (adress >> 0 ) & 0xff;

	rx_cnt=8;		// 8 Bytes sollen empfangen werden
	if(DapiTCPSendCommand(handle, tx_buffer, tx_cnt, rx, rx_cnt)) 
	{
		#ifdef __linux__
			data =((ULONGLONG) (rx[7]) << 56) & 0xff00000000000000LL;
			data|=((ULONGLONG) (rx[6]) << 48) & 0x00ff000000000000LL;
			data|=((ULONGLONG) (rx[5]) << 40) & 0x0000ff0000000000LL;
			data|=((ULONGLONG) (rx[4]) << 32) & 0x000000ff00000000LL;
			data|=((ULONGLONG) (rx[3]) << 24) & 0x00000000ff000000LL;
			data|=((ULONGLONG) (rx[2]) << 16) & 0x0000000000ff0000LL;
			data|=((ULONGLONG) (rx[1]) << 8 ) & 0x000000000000ff00LL;
			data|=((ULONGLONG) (rx[0]) << 0 ) & 0x00000000000000ffLL;
		#else
			data =((ULONGLONG) (rx[7]) << 56) & 0xff00000000000000;
			data|=((ULONGLONG) (rx[6]) << 48) & 0x00ff000000000000;
			data|=((ULONGLONG) (rx[5]) << 40) & 0x0000ff0000000000;
			data|=((ULONGLONG) (rx[4]) << 32) & 0x000000ff00000000;
			data|=((ULONGLONG) (rx[3]) << 24) & 0x00000000ff000000;
			data|=((ULONGLONG) (rx[2]) << 16) & 0x0000000000ff0000;
			data|=((ULONGLONG) (rx[1]) << 8 ) & 0x000000000000ff00;
			data|=((ULONGLONG) (rx[0]) << 0 ) & 0x00000000000000ff;
		#endif
		
		return data;
	}
	else 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
		return(0L);
	}
}


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void DapiTCPSpecialCmd(unsigned long handle,unsigned long cmd)
{
	UCHAR tx_buffer[256];
	unsigned long tx_cnt=0;

	tx_buffer[tx_cnt++]='Z';		// SecialCmd
	tx_buffer[tx_cnt++]='C';		// SecialCmd

	tx_buffer[tx_cnt++]=(UCHAR) (cmd >> 8)  & 0xff;
	tx_buffer[tx_cnt++]=(UCHAR) (cmd >> 0 ) & 0xff;
	
	if(!DapiTCPSendSpecialCommand(handle, tx_buffer, tx_cnt, 0, 0)) 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiTCPWriteMultipleBytesOLD(ULONG handle, ULONG adress, ULONG adress_tiefe, ULONG wiederholungen, unsigned char* data, ULONG data_length)
{

	ULONG adr;
	ULONG anz;
	ULONG pos=0;
	ULONGLONG d64;

	for(anz=0;anz!=wiederholungen;++anz)
	{
		adr=adress;
		
		do 
		{
			if((adr+8) <= (adress+adress_tiefe))
			{
				// 8 Byte

				d64 =  (ULONGLONG) *(data+pos+0);
				d64|= ((ULONGLONG) *(data+pos+1)) << 8;
				d64|= ((ULONGLONG) *(data+pos+2)) << 16;
				d64|= ((ULONGLONG) *(data+pos+3)) << 24;
				d64|= ((ULONGLONG) *(data+pos+4)) << 32;
				d64|= ((ULONGLONG) *(data+pos+5)) << 40;
				d64|= ((ULONGLONG) *(data+pos+6)) << 48;
				d64|= ((ULONGLONG) *(data+pos+7)) << 56;

				DapiWriteLongLong(handle, adr, d64);
				adr+=8;
				pos+=8;
			}
			else
			{
				DapiWriteByte(handle, adr, *(data+pos));
				adr++;
				pos++;
			}
		} while(adr < (adress+adress_tiefe));
	}
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiTCPWriteMultipleBytes(ULONG handle, ULONG address, ULONG address_depth, ULONG increment, unsigned char *data, ULONG data_length)
{
	UCHAR *tx_buffer;
	char debug_msg[200];
	
	unsigned long tx_cnt;

	ULONG data_id = 0;
	ULONG i;
	ULONG j;
	ULONG pos;
	
	tx_cnt=0;
	pos=0;
	
	// speicherallokierung fr tx_buffer + check 
	tx_buffer = (UCHAR *) malloc((((address_depth + 3) * increment) + TCP_RO_BIG_TX_HEADER_LENGTH));	// header(8) + wiederholung * (adresstiefe + start_id(1) + data_id(2))
	
	if (tx_buffer == NULL)
	{
		sprintf(debug_msg, "DapiTCPWriteMultipleBytes / malloc() for tx_buffer failed");
		DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

		return 1;			// Error
	}
	
	// tx_buffer ok - daten fllen
	tx_buffer[tx_cnt++]='W';	// COMMAND //Write
	tx_buffer[tx_cnt++]='M';	// WIDTH //Multiple

	// address (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((address >> 8 ) & 0xff);	// ADDRESS_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((address >> 0 ) & 0xff);	// ADDRESS_BIT_0_7
	
	// address_depth (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((address_depth >> 8 ) & 0xff);	// ADDRESS_DEPTH_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((address_depth >> 0 ) & 0xff);	// ADDRESS_DEPTH_BIT_0_7
	
	// increment (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((increment >> 8 ) & 0xff);	// INCREMENT_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((increment >> 0 ) & 0xff);	// INCREMENT_BIT_0_7
	
	// data start
	for (i=0; i<increment; i++)
	{
		// START_ID_MULTIPLE_BYTE (1byte)
		tx_buffer[tx_cnt++]= DEDITEC_TCP_START_ID_FOR_MULTIPLE_BYTE_DATA;		// START_ID_MULTIPLE_BYTE
		
		// DATA_ID (2byte)
		tx_buffer[tx_cnt++]= (unsigned char) ((i >> 8 ) & 0xff);			// DATA_ID_BIT_8_15
		tx_buffer[tx_cnt++]= (unsigned char) ((i >> 0 ) & 0xff);			// DATA_ID_BIT_0_7
		
		// data
		for (j=0; j<address_depth; j++)
		{
			tx_buffer[tx_cnt++]= (*(data + (pos++)));
		}
	}
	// data end
	
	if(!DapiTCPSendCommand(handle, tx_buffer, tx_cnt, 0, 0)) 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}

	// speicherplatz von tx_buffer wieder freigeben
	free(tx_buffer);
	
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiTCPReadMultipleBytes(ULONG handle, ULONG address, ULONG address_depth, ULONG increment, unsigned char *data, ULONG data_length)
{
	char debug_msg[200];

	UCHAR tx_buffer[2000];
	UCHAR *data_temp;
	ULONG data_temp_length;
	
	ULONG i;
	ULONG j;
	ULONG pos;
	ULONG tx_cnt;
	
	ULONG m_start_id;
	ULONG m_data_id;
	ULONG m_data_pos;

	DAPI_HANDLE *DapiHandle;

	DapiHandle = (DAPI_HANDLE*) handle;
	
	// check ob der "richtige buffer" gro genug ist	
	if (data_length < (increment * address_depth))
	{
		sprintf(debug_msg, "DapiTCPReadMultipleBytes / Buffer too small - address_depth=%ld * increment=%ld < data_length=%ld", address_depth, increment, data_length);
		DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

		return 1;			// Error
	}	

	// speicherallokierung fr data_temp + check 
	data_temp = (UCHAR *) malloc((((address_depth + 3) * increment) + 1));	// +1 weil "sizeof" nicht bei 0 sondern bei 1 anfngt
	
	if (data_temp == NULL)
	{
		sprintf(debug_msg, "DapiTCPReadMultipleBytes / malloc() for temp_buffer failed");
		DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

		return 1;			// Error
	}
	
	data_temp_length = ((address_depth + 3) * increment);//sizeof(data_temp);
	
	// tx_buffer zusammenbauen
	tx_cnt=0;
	pos=0;

	tx_buffer[tx_cnt++]='R';	// COMMAND //Write
	tx_buffer[tx_cnt++]='M';	// WIDTH //Multiple

	// address (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((address >> 8 ) & 0xff);	// ADDRESS_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((address >> 0 ) & 0xff);	// ADDRESS_BIT_0_7
	
	// address_depth (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((address_depth >> 8 ) & 0xff);	// ADDRESS_DEPTH_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((address_depth >> 0 ) & 0xff);	// ADDRESS_DEPTH_BIT_0_7
	
	// increment (2byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((increment >> 8 ) & 0xff);	// INCREMENT_BIT_8_15
	tx_buffer[tx_cnt++]= (unsigned char) ((increment >> 0 ) & 0xff);	// INCREMENT_BIT_0_7
	
	// anfrage an ETH-modul schicken
	if(!DapiTCPSendCommand(handle, tx_buffer, tx_cnt, data_temp, data_temp_length)) 
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);					
	}
	
	// data_temp berprfen
	// datenblock = start_id_multiple_byte(1) +  data_id(2) + addr_depth	

	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTCPReadMultipleBytes");

		sprintf(debug_msg,"RECEIVING %lu bytes:", data_temp_length);
		debug_print(debug_msg);

		debug_hexdump_buffer_to_string(data_temp, data_temp_length, debug_msg, sizeof(debug_msg));
		debug_print(debug_msg);
	#endif
	// -----
	
	for (i=0; i<increment; i++)
	{
		pos = 0;
		m_data_pos = i*(address_depth+3);
		
		// start_id + data_id aus dem buffer filtern
		m_start_id = *(data_temp + m_data_pos + (pos++));
		m_data_id = *(data_temp + m_data_pos + (pos++));
		m_data_id = (m_data_id << 8) | *(data_temp + m_data_pos + (pos++));
		
		// start_id checken
		if (m_start_id != DEDITEC_TCP_START_ID_FOR_MULTIPLE_BYTE_DATA)
		{
			sprintf(debug_msg, "DapiTCPReadMultipleBytes / Wrong START_ID");
			DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

			return 1;			// Error
		}
		
		// data_id checken
		if (m_data_id != i)
		{	
			sprintf(debug_msg, "DapiTCPReadMultipleBytes / Wrong DATA_ID");
			DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

			return 1;			// Error
		}
	}
	
	// daten sind ok - daten in "richtigen buffer" schreiben
	tx_cnt = 0;
	
	for (i=0; i< increment; i++)
	{
		for (j=0; j<address_depth; j++)
		{
			*(data + (tx_cnt++)) = *(data_temp + 3 + j + i*(address_depth+3));
		}
	}
	
	// zum abschluss geschrieben zeichen in "richtigen buffer" mit der errechneten (und bentigten gre vergleichen)
	if (tx_cnt != (increment * address_depth))
	{
		sprintf(debug_msg, "DapiTCPReadMultipleBytes / Buffer copy error");
		DapiSetError(DAPI_ERR_WITH_TEXT, 0, debug_msg, __FILE__, __LINE__);					

		return 1;			// Error
	}
	
	// DEBUG
	#ifdef DEBUG_OUTPUT
		debug_set_function_name("DapiTCPReadMultipleBytes");

		sprintf(debug_msg,"RETURNING %lu bytes:", data_temp_length);
		debug_print(debug_msg);

		debug_hexdump_buffer_to_string(data, data_length, debug_msg, sizeof(debug_msg));
		debug_print(debug_msg);
	#endif
	// -----
	
	// speicherplatz von data_temp wieder freigeben
	free(data_temp);	
	
	return 0;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

ULONG DapiTCPReadFifo(ULONG handle, ULONG type, unsigned char *data, ULONG data_length)
{
	UCHAR tx_buffer[2000];
	ULONG pos;
	ULONG tx_cnt;
	ULONG rx_data_length;

	// tx_buffer zusammenbauen
	tx_cnt=0;
	pos=0;

	tx_buffer[tx_cnt++]='R';	// COMMAND //Write
	tx_buffer[tx_cnt++]='F';	// WIDTH //Multiple

	// type (4byte)
	tx_buffer[tx_cnt++]= (unsigned char) ((type >> 24) & 0xff);	// type
	tx_buffer[tx_cnt++]= (unsigned char) ((type >> 16) & 0xff);	// type
	tx_buffer[tx_cnt++]= (unsigned char) ((type >> 8 ) & 0xff);	// type
	tx_buffer[tx_cnt++]= (unsigned char) ((type >> 0 ) & 0xff);	// type
	
	// Anfrage an ETH-modul schicken
	if(!DapiTCPSendCommandBig(handle, tx_buffer, tx_cnt, data, data_length, &rx_data_length))
	{
		DapiSetError(DAPI_ERR_COMMUNICATION_ERROR, 0, NULL, __FILE__, __LINE__);
		rx_data_length=0;
	}

	return rx_data_length;
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

