//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//
//
//
//	deditec_debug_comfort.c
//	project: delib
//
//
//	(c) DEDITEC GmbH, 2013
//	web: http://www.deditec.de/
//	mail: vertrieb@deditec.de
//
// 
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************
//***********************************************************************************

// ------------------------------------
// ------------------------------------
// History
// ------------------------------------
// ------------------------------------
//
// DT	26.11.2013	Start DEDITEC-History
// CC	17.12.2013	Added function debug_sprintf and debug_strcat



// ------------------------------------
// ------------------------------------
// Functions
// ------------------------------------
// ------------------------------------

// debug_print
// debug_info
// debug_error
// ------------------------------------
// debug_sprintf
// debug_strcat
// ------------------------------------
// debug_print_hex
// debug_strcat_hex

// ------------------------------------
// ------------------------------------
// Includes
// ------------------------------------
// ------------------------------------

#ifdef _MSC_VER
#include <windows.h>	// OutputDebugString
#endif

#include <stdarg.h>		// va_list(), va_start(), va_end()
#include <stdio.h>		// vsprintf()

#ifdef DELIB_COMPILE
#include "../../proj_config.h"
#endif

// ------------------------------------
// ------------------------------------
// Global Vars
// ------------------------------------
// ------------------------------------

char debug_msg[200];

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_print(const char * format, ...)
{
	#ifdef debug_version
		va_list args;

		va_start(args, format);

		#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
			vsprintf_s(debug_msg, sizeof(debug_msg), format, args);
		#else
			vsprintf(debug_msg, format, args);
		#endif

		va_end(args);

		debug_out_print(debug_msg);
	#else
		return;
	#endif
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_info(const char * format, ...)
{
	va_list args;

	va_start(args, format);
	
	#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
		vsprintf_s(debug_msg, sizeof(debug_msg), format, args);
	#else
		vsprintf(debug_msg, format, args);
	#endif

	va_end(args);

	debug_out_info(debug_msg);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_error(const char * format, ...)
{
	va_list args;

	va_start(args, format);	

	#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
		vsprintf_s(debug_msg, sizeof(debug_msg), format, args);
	#else
		vsprintf(debug_msg, format, args);
	#endif

	va_end(args);

	debug_out_error(debug_msg);
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_print_hex(unsigned char *src_buffer, unsigned long src_buffer_len)
{
	#ifdef debug_version
		if (debug_hexdump_buffer_to_string(src_buffer, src_buffer_len, debug_msg, sizeof(debug_msg)) == RETURN_OK)
		{
			debug_print(debug_msg);
		}
	#else
		return;
	#endif
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_sprintf(char *dest_buffer, const char *format, ...)
{
	#ifdef debug_version
		va_list args;

		va_start(args, format);

		#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
			vsprintf_s(debug_msg, sizeof(debug_msg), format, args);
		#else
			vsprintf(debug_msg, format, args);
		#endif

		va_end(args);
	#else
		return;
	#endif
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_strcat(char *dest_buffer, const char *format, ...)
{
	#ifdef debug_version
		va_list args;

		va_start(args, format);

		#if (_MSC_VER >= 1500)	// >= Visual Studio 2008
			vsprintf_s(debug_msg, sizeof(debug_msg), format, args);
		#else
			vsprintf(debug_msg, format, args);
		#endif

		va_end(args);
		
		strcat_n(dest_buffer, sizeof(dest_buffer), debug_msg, strlen_n(debug_msg, sizeof(debug_msg)));
	#else
		return;
	#endif
}

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

void debug_strcat_hex(char *dest_buffer, unsigned long dest_buffer_len, unsigned char *src_buffer, unsigned long src_buffer_len)
{
	#ifdef debug_version
		if (debug_hexdump_buffer_to_string(src_buffer, src_buffer_len, debug_msg, sizeof(debug_msg)) == RETURN_OK)
		{
			strcat_n(dest_buffer, dest_buffer_len, debug_msg, strlen_n(debug_msg, sizeof(debug_msg)));
		}
	#else
		return;
	#endif
}
